import { reflect } from 'kaop';
import 'reflect-metadata';

const KEY_ORIGINAL_METHOD = "original-method";
const KEY_BEFORE_METHOD = "before-method";
const KEY_AFTER_METHOD = "after-method";
const KEY_BEFORE_INSTANCE = "before-instance";
const KEY_AFTER_INSTANCE = "after-instance";

function generateKey(scope, methodName) {
    return `${scope}-${methodName}`;
}
function wrapMethod(target, methodName, beforeKey, afterKey, caller) {
    const keyOriginalMethod = generateKey(KEY_ORIGINAL_METHOD, methodName);
    return reflect.createProxyFn(target, methodName, [
        ...Reflect.getMetadata(beforeKey, target) || [],
        Reflect.getMetadata(keyOriginalMethod, target),
        ...Reflect.getMetadata(afterKey, target) || []
    ], caller);
}
function applyReflect(target, advices, methodName, keyJoinPoint, original) {
    const keyOriginalMethod = generateKey(KEY_ORIGINAL_METHOD, methodName);
    const adviceArr = Reflect.getMetadata(keyJoinPoint, target) || [];
    adviceArr.push(...advices.map(reflect.advice));
    Reflect.defineMetadata(keyJoinPoint, adviceArr, target);
    if (!Reflect.getMetadata(keyOriginalMethod, target))
        Reflect.defineMetadata(keyOriginalMethod, original, target);
    if (methodName === "constructor") {
        const keyBeforeInstance = generateKey(KEY_BEFORE_INSTANCE, methodName);
        const keyAfterInstance = generateKey(KEY_AFTER_INSTANCE, methodName);
        const result = wrapMethod(target, methodName, keyBeforeInstance, keyAfterInstance, meta => Reflect.construct(meta.target, meta.args, meta.ES6newTarget));
        result.prototype = target.prototype;
        return result;
    }
    else {
        const keyBeforeMethod = generateKey(KEY_BEFORE_METHOD, methodName);
        const keyAfterMethod = generateKey(KEY_AFTER_METHOD, methodName);
        return wrapMethod(target, methodName, keyBeforeMethod, keyAfterMethod);
    }
}
function beforeMethod(...advices) {
    return function (target, methodName, descriptor) {
        const keyBeforeMethod = generateKey(KEY_BEFORE_METHOD, methodName);
        descriptor.value = applyReflect(target, advices, methodName, keyBeforeMethod, descriptor.value);
        return descriptor;
    };
}
function afterMethod(...advices) {
    return function (target, methodName, descriptor) {
        const keyAfterMethod = generateKey(KEY_AFTER_METHOD, methodName);
        descriptor.value = applyReflect(target, advices, methodName, keyAfterMethod, descriptor.value);
        return descriptor;
    };
}
function beforeInstance(...advices) {
    return function (target, methodName = "constructor") {
        const keyBeforeInstance = generateKey(KEY_BEFORE_INSTANCE, methodName);
        return applyReflect(target, advices, methodName, keyBeforeInstance, target);
    };
}
function afterInstance(...advices) {
    return function (target, methodName = "constructor") {
        const keyAfterInstance = generateKey(KEY_AFTER_INSTANCE, methodName);
        return applyReflect(target, advices, methodName, keyAfterInstance, target);
    };
}

function onException(advice) {
    return afterMethod(meta => meta.exception && advice(meta));
}

function applyAspect(definition) {
    return function (target) {
        const ctor = definition["constructor"] instanceof Array ? definition["constructor"] : [];
        delete definition["constructor"];
        for (let key in definition) {
            definition[key].forEach(advice => Object.defineProperty(target.prototype, key, advice(target, key, Object.getOwnPropertyDescriptor(target.prototype, key))));
        }
        ctor.forEach(advice => target = advice(target));
        return target;
    };
}

export { afterMethod, beforeMethod, afterInstance, beforeInstance, onException, applyAspect };
